#
# Copyright 2014 the original author or authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import pkg_resources
import mock

from cabalgata.silla import factories
from cabalgata.silla.util import disk


def plugin_entry_points():
    return (
        pkg_resources.EntryPoint.parse('a=data.factories:FactoryA'),
    )


@mock.patch('pkg_resources.iter_entry_points')
def test_load_factory(mock_pkg_resources):
    mock_pkg_resources.return_value = plugin_entry_points()

    with disk.temp_directory() as temp_directory:
        factory = factories.load_factory('a', temp_directory)

        assert factory.versions
        assert factory.definitions

        try:
            factories.load_factory('does_not_exist', temp_directory)

            assert False, 'Should have raised an KeyError'
        except KeyError:
            pass


@mock.patch('pkg_resources.iter_entry_points')
def test_install(mock_pkg_resources):
    mock_pkg_resources.return_value = plugin_entry_points()

    with disk.temp_directory() as temp_directory:
        factory = factories.load_factory('a', temp_directory)

        assert not factory.installed

        factory.install('test', '1.2.3')

        installed = factory.installed
        assert installed['test'] == '1.2.3'

        p = factory.load('test')

        assert p.version == '1.2.3'

        factory.uninstall('test')

        assert not factory.installed


@mock.patch('pkg_resources.iter_entry_points')
def test_start(mock_pkg_resources):
    mock_pkg_resources.return_value = plugin_entry_points()

    with disk.temp_directory() as temp_directory:
        factory = factories.load_factory('a', temp_directory)

        factory.install('test', '1.2.3')

        p = factory.load('test')
        assert not p.running

        p.start()
        assert p.running
        p.stop()
        assert not p.running

        p.start()
        assert p.running
        p.kill()
        assert not p.running

        factory.uninstall('test')



#
# Copyright 2014 the original author or authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from distutils import version

from cabalgata.silla import catalog
from cabalgata.silla.util import disk


def test_catalog():
    json = {'version': '1.2.3'}
    c = catalog.Catalog.from_json(json)

    assert c == catalog.Catalog('1.2.3')
    assert c.to_json() == json


def test_initialize():
    with disk.temp_directory() as temp_dir:
        catalog.initialize(catalog_dir=temp_dir)

        with catalog.load_catalog(temp_dir) as c:
            assert c.version
            assert version.StrictVersion(c.version)



#
# Copyright 2014 the original author or authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
""" Disk utilities
"""
import contextlib
import os
import shutil
import tempfile
import errno


@contextlib.contextmanager
def temp_directory(*args, **kwargs):
    """
    Context manager returns a path created by mkdtemp and cleans it up afterwards.
    """

    path = tempfile.mkdtemp(*args, **kwargs)
    try:
        yield path
    finally:
        shutil.rmtree(path)


def make_directories(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



#
# Copyright 2014 the original author or authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
""" Catalog management functions
"""
import contextlib
import json
import logging
import os

from cabalgata import silla
from cabalgata.silla.util import disk


CATALOG_DIR = '.cabalgata'
CATALOG_FILE = 'catalog.json'

log = logging.getLogger(__name__)


class CorruptedCatalogError(Exception):
    """Error for corrupted catalog."""


class Catalog(object):
    def __init__(self, version):
        self.version = version

    def as_tuple(self):
        return self.version,

    def __eq__(self, other):
        return self.as_tuple() == other.as_tuple()

    def __repr__(self):
        return 'Catalog(%r)' % self.as_tuple()

    @classmethod
    def from_json(cls, json):
        return cls(json['version'])

    def to_json(self):
        return {
            'version': self.version
        }


def initialize(catalog_dir=None):
    catalog_dir = catalog_dir or os.path.join(os.path.abspath('~'), CATALOG_DIR)

    if not os.path.exists(catalog_dir):
        disk.make_directories(catalog_dir)
        with load_catalog(catalog_dir):
            pass


@contextlib.contextmanager
def load_catalog(catalog_dir):
    catalog_file = os.path.join(catalog_dir, CATALOG_FILE)
    if os.path.exists(catalog_file):
        try:
            with open(catalog_file) as fp:
                json_catalog = json.load(fp)
                catalog = Catalog.from_json(json_catalog)
        except Exception:
            log.exception('Unable to read %s' % catalog_file)
            raise CorruptedCatalogError('Unable to read %s' % catalog_file)
    else:
        catalog = Catalog(silla.VERSION)

    yield catalog

    try:
        json_catalog = catalog.to_json()
        with open(catalog_file, 'w') as fp:
            json.dump(json_catalog, fp)
    except Exception:
        log.exception('Unable to write %s' % catalog_file)
        raise CorruptedCatalogError('Unable to write %s' % catalog_file)



#
# Copyright 2014 the original author or authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import os
import errno
import time


class TimeoutExpired(Exception):
    pass


def pid_exists(pid):
    """Check whether pid exists in the current process table."""
    if pid < 0:
        return False
    try:
        os.kill(pid, 0)
    except OSError as e:
        return e.errno == errno.EPERM
    else:
        return True


def wait_pid(pid, timeout=None):
    """Wait for process with pid 'pid' to terminate and return its
    exit status code as an integer.

    If pid is not a children of os.getpid() (current process) just
    waits until the process disappears and return None.

    If pid does not exist at all return None immediately.

    Raise TimeoutExpired on timeout expired (if specified).
    """

    def check_timeout(delay):
        if timeout is not None:
            if time.time() >= stop_at:
                raise TimeoutExpired
        time.sleep(delay)
        return min(delay * 2, 0.04)

    if timeout is not None:
        waitcall = lambda: os.waitpid(pid, os.WNOHANG)
        stop_at = time.time() + timeout
    else:
        waitcall = lambda: os.waitpid(pid, 0)

    delay = 0.0001
    while 1:
        try:
            retpid, status = waitcall()
        except OSError as err:
            if err.errno == errno.EINTR:
                delay = check_timeout(delay)
                continue
            elif err.errno == errno.ECHILD:
                # This has two meanings:
                # - pid is not a child of os.getpid() in which case
                # we keep polling until it's gone
                # - pid never existed in the first place
                # In both cases we'll eventually return None as we
                # can't determine its exit status code.
                while 1:
                    if pid_exists(pid):
                        delay = check_timeout(delay)
                    else:
                        return
            else:
                raise
        else:
            if retpid == 0:
                # WNOHANG was used, pid is still running
                delay = check_timeout(delay)
                continue
            # process exited due to a signal; return the integer of
            # that signal
            if os.WIFSIGNALED(status):
                return os.WTERMSIG(status)
            # process exited using exit(2) system call; return the
            # integer exit(2) system call has been called with
            elif os.WIFEXITED(status):
                return os.WEXITSTATUS(status)
            else:
                # should never happen
                raise RuntimeError("unknown process exit status")



#
# Copyright 2014 the original author or authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
""" Configuration of installation and service
"""

NO_DEFAULT = object()


class Definition(object):
    def __init__(self, name, kind, secret, default=NO_DEFAULT):
        self.name = name
        self.kind = kind
        self.secret = secret
        self.default = default

    def as_tuple(self):
        return self.name, self.kind, self.secret, self.default

    def __eq__(self, other):
        return self.as_tuple() == other.as_tuple()

    def __repr__(self):
        return 'Definition(%r, %r, %r, %r)' % self.as_tuple()
